@@ -3,11 +3,12 @@
 //! The virtual file system abstraction generalizes over file systems and allow using
 //! different VirtualFileSystem implementations (i.e. an in memory implementation for unit tests)
 
-use crate::error::VfsResultExt;
-use crate::{FileSystem, VfsError, VfsResult};
 use std::io::{Read, Seek, Write};
 use std::sync::Arc;
 
+use crate::error::VfsResultExt;
+use crate::{FileSystem, VfsError, VfsResult};
+
 /// Trait combining Seek and Read, return value for opening files
 pub trait SeekAndRead: Seek + Read {}
 
@@ -53,6 +54,11 @@ impl Eq for VfsPath {}
 
 impl VfsPath {
     /// Creates a root path for the given filesystem
+    ///
+    /// ```
+    /// # use vfs::{PhysicalFS, VfsPath};
+    /// let path = VfsPath::new(PhysicalFS::new("."));
+    /// ````
     pub fn new<T: FileSystem>(filesystem: T) -> Self {
         VfsPath {
             path: "".to_string(),
@@ -63,14 +69,38 @@ impl VfsPath {
     }
 
     /// Returns the string representation of this path
+    ///
+    /// ```
+    /// # use vfs::{PhysicalFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(PhysicalFS::new("."));
+    ///
+    /// assert_eq!(path.as_str(), "");
+    /// assert_eq!(path.join("foo.txt")?.as_str(), "/foo.txt");
+    /// # Ok::<(), VfsError>(())
+    /// ````
     pub fn as_str(&self) -> &str {
         &self.path
     }
 
     /// Appends a path segment to this path, returning the result
+    ///
+    /// ```
+    /// # use vfs::{PhysicalFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(PhysicalFS::new("."));
+    ///
+    /// assert_eq!(path.join("foo.txt")?.as_str(), "/foo.txt");
+    /// assert_eq!(path.join("foo/bar.txt")?.as_str(), "/foo/bar.txt");
+    ///
+    /// let foo = path.join("foo")?;
+    ///
+    /// assert_eq!(path.join("foo/bar.txt")?, foo.join("bar.txt")?);
+    /// assert_eq!(path, foo.join("..")?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn join(&self, path: impl AsRef<str>) -> VfsResult<Self> {
         self.join_internal(path.as_ref())
     }
+
     /// Appends a path segment to this path, returning the result
     fn join_internal(&self, path: &str) -> VfsResult<Self> {
         if path.is_empty() {
@@ -111,25 +141,39 @@ impl VfsPath {
             fs: self.fs.clone(),
         })
     }
-    /// Iterates over all entries of this directory path
-    pub fn read_dir(&self) -> VfsResult<Box<dyn Iterator<Item = VfsPath>>> {
-        let parent = self.path.clone();
-        let fs = self.fs.clone();
-        Ok(Box::new(
-            self.fs
-                .fs
-                .read_dir(&self.path)
-                .with_context(|| format!("Could not read directory '{}'", &self.path))?
-                .map(move |path| VfsPath {
-                    path: format!("{}/{}", parent, path),
-                    fs: fs.clone(),
-                }),
-        ))
+
+    /// Returns the root path of this filesystem
+    ///
+    /// ```
+    /// # use vfs::{MemoryFS, VfsError, VfsFileType, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo/bar")?;
+    ///
+    /// assert_eq!(directory.root(), path);
+    /// # Ok::<(), VfsError>(())
+    /// ```
+    pub fn root(&self) -> VfsPath {
+        VfsPath {
+            path: "".to_string(),
+            fs: self.fs.clone(),
+        }
     }
 
     /// Creates the directory at this path
     ///
     /// Note that the parent directory must exist.
+    ///
+    /// ```
+    /// # use vfs::{MemoryFS, VfsError, VfsFileType, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    ///
+    /// directory.create_dir()?;
+    ///
+    /// assert!(directory.exists()?);
+    /// assert_eq!(directory.metadata()?.file_type, VfsFileType::Directory);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn create_dir(&self) -> VfsResult<()> {
         self.get_parent("create directory")?;
         self.fs
@@ -139,6 +183,21 @@ impl VfsPath {
     }
 
     /// Creates the directory at this path, also creating parent directories as necessary
+    ///
+    /// ```
+    /// # use vfs::{MemoryFS, VfsError, VfsFileType, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo/bar")?;
+    ///
+    /// directory.create_dir_all()?;
+    ///
+    /// assert!(directory.exists()?);
+    /// assert_eq!(directory.metadata()?.file_type, VfsFileType::Directory);
+    /// let parent = path.join("foo")?;
+    /// assert!(parent.exists()?);
+    /// assert_eq!(parent.metadata()?.file_type, VfsFileType::Directory);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn create_dir_all(&self) -> VfsResult<()> {
         let mut pos = 1;
         let path = &self.path;
@@ -164,15 +223,50 @@ impl VfsPath {
         Ok(())
     }
 
-    /// Opens the file at this path for reading
-    pub fn open_file(&self) -> VfsResult<Box<dyn SeekAndRead>> {
-        self.fs
-            .fs
-            .open_file(&self.path)
-            .with_context(|| format!("Could not open file '{}'", &self.path))
+    /// Iterates over all entries of this directory path
+    ///
+    /// ```
+    /// # use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// path.join("foo")?.create_dir()?;
+    /// path.join("bar")?.create_dir()?;
+    ///
+    /// let mut directories: Vec<_> = path.read_dir()?.collect();
+    ///
+    /// directories.sort_by_key(|path| path.as_str().to_string());
+    /// assert_eq!(directories, vec![path.join("bar")?, path.join("foo")?]);
+    /// # Ok::<(), VfsError>(())
+    /// ```
+    pub fn read_dir(&self) -> VfsResult<Box<dyn Iterator<Item = VfsPath>>> {
+        let parent = self.path.clone();
+        let fs = self.fs.clone();
+        Ok(Box::new(
+            self.fs
+                .fs
+                .read_dir(&self.path)
+                .with_context(|| format!("Could not read directory '{}'", &self.path))?
+                .map(move |path| VfsPath {
+                    path: format!("{}/{}", parent, path),
+                    fs: fs.clone(),
+                }),
+        ))
     }
 
     /// Creates a file at this path for writing, overwriting any existing file
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo.txt")?;
+    ///
+    /// write!(file.create_file()?, "Hello, world!")?;
+    ///
+    /// let mut result = String::new();
+    /// file.open_file()?.read_to_string(&mut result)?;
+    /// assert_eq!(&result, "Hello, world!");
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn create_file(&self) -> VfsResult<Box<dyn Write>> {
         self.get_parent("create file")?;
         self.fs
@@ -181,6 +275,28 @@ impl VfsPath {
             .with_context(|| format!("Could not create file '{}'", &self.path))
     }
 
+    /// Opens the file at this path for reading
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo.txt")?;
+    /// write!(file.create_file()?, "Hello, world!")?;
+    /// let mut result = String::new();
+    ///
+    /// file.open_file()?.read_to_string(&mut result)?;
+    ///
+    /// assert_eq!(&result, "Hello, world!");
+    /// # Ok::<(), VfsError>(())
+    /// ```
+    pub fn open_file(&self) -> VfsResult<Box<dyn SeekAndRead>> {
+        self.fs
+            .fs
+            .open_file(&self.path)
+            .with_context(|| format!("Could not open file '{}'", &self.path))
+    }
+
     /// Checks whether parent is a directory
     fn get_parent(&self, action: &str) -> VfsResult<()> {
         let parent = self.parent();
@@ -214,6 +330,19 @@ impl VfsPath {
     }
 
     /// Opens the file at this path for appending
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo.txt")?;
+    /// write!(file.create_file()?, "Hello, ")?;
+    /// write!(file.append_file()?, "world!")?;
+    /// let mut result = String::new();
+    /// file.open_file()?.read_to_string(&mut result)?;
+    /// assert_eq!(&result, "Hello, world!");
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn append_file(&self) -> VfsResult<Box<dyn Write>> {
         self.fs
             .fs
@@ -222,6 +351,20 @@ impl VfsPath {
     }
 
     /// Removes the file at this path
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo.txt")?;
+    /// write!(file.create_file()?, "Hello, ")?;
+    /// assert!(file.exists()?);
+    ///
+    /// file.remove_file()?;
+    ///
+    /// assert!(!file.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn remove_file(&self) -> VfsResult<()> {
         self.fs
             .fs
@@ -232,6 +375,20 @@ impl VfsPath {
     /// Removes the directory at this path
     ///
     /// The directory must be empty.
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    /// directory.create_dir();
+    /// assert!(directory.exists()?);
+    ///
+    /// directory.remove_dir()?;
+    ///
+    /// assert!(!directory.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn remove_dir(&self) -> VfsResult<()> {
         self.fs
             .fs
@@ -242,6 +399,20 @@ impl VfsPath {
     /// Ensures that the directory at this path is removed, recursively deleting all contents if necessary
     ///
     /// Returns successfully if directory does not exist
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    /// directory.join("bar")?.create_dir_all();
+    /// assert!(directory.exists()?);
+    ///
+    /// directory.remove_dir_all()?;
+    ///
+    /// assert!(!directory.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn remove_dir_all(&self) -> VfsResult<()> {
         if !self.exists()? {
             return Ok(());
@@ -258,6 +429,23 @@ impl VfsPath {
     }
 
     /// Returns the file metadata for the file at this path
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    /// directory.create_dir();
+    ///
+    /// assert_eq!(directory.metadata()?.len, 0);
+    /// assert_eq!(directory.metadata()?.file_type, VfsFileType::Directory);
+    ///
+    /// let file = path.join("bar.txt")?;
+    /// write!(file.create_file()?, "Hello, world!")?;
+    ///
+    /// assert_eq!(file.metadata()?.len, 13);
+    /// assert_eq!(file.metadata()?.file_type, VfsFileType::File);
+    /// # Ok::<(), VfsError>(())
     pub fn metadata(&self) -> VfsResult<VfsMetadata> {
         self.fs
             .fs
@@ -265,18 +453,100 @@ impl VfsPath {
             .with_context(|| format!("Could not get metadata for '{}'", &self.path))
     }
 
+    /// Returns `true` if the path exists and is pointing at a regular file, otherwise returns `false`.
+    ///
+    /// Note that this call may fail if the file's existence cannot be determined or the metadata can not be retrieved
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    /// directory.create_dir()?;
+    /// let file = path.join("foo.txt")?;
+    /// file.create_file()?;
+    ///
+    /// assert!(!directory.is_file()?);
+    /// assert!(file.is_file()?);
+    /// # Ok::<(), VfsError>(())
+    pub fn is_file(&self) -> VfsResult<bool> {
+        if !self.exists()? {
+            return Ok(false);
+        }
+        let metadata = self.metadata()?;
+        Ok(metadata.file_type == VfsFileType::File)
+    }
+
+    /// Returns `true` if the path exists and is pointing at a directory, otherwise returns `false`.
+    ///
+    /// Note that this call may fail if the directory's existence cannot be determined or the metadata can not be retrieved
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    /// directory.create_dir()?;
+    /// let file = path.join("foo.txt")?;
+    /// file.create_file()?;
+    ///
+    /// assert!(directory.is_dir()?);
+    /// assert!(!file.is_dir()?);
+    /// # Ok::<(), VfsError>(())
+    pub fn is_dir(&self) -> VfsResult<bool> {
+        if !self.exists()? {
+            return Ok(false);
+        }
+        let metadata = self.metadata()?;
+        Ok(metadata.file_type == VfsFileType::Directory)
+    }
+
     /// Returns true if a file or directory exists at this path, false otherwise
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let directory = path.join("foo")?;
+    ///
+    /// assert!(!directory.exists()?);
+    ///
+    /// directory.create_dir();
+    ///
+    /// assert!(directory.exists()?);
+    /// # Ok::<(), VfsError>(())
     pub fn exists(&self) -> VfsResult<bool> {
         self.fs.fs.exists(&self.path)
     }
 
     /// Returns the filename portion of this path
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo/bar.txt")?;
+    ///
+    /// assert_eq!(&file.filename(), "bar.txt");
+    ///
+    /// # Ok::<(), VfsError>(())
     pub fn filename(&self) -> String {
         let index = self.path.rfind('/').map(|x| x + 1).unwrap_or(0);
         self.path[index..].to_string()
     }
 
     /// Returns the extension portion of this path
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    ///
+    /// assert_eq!(path.join("foo/bar.txt")?.extension(), Some("txt".to_string()));
+    /// assert_eq!(path.join("foo/bar.txt.zip")?.extension(), Some("zip".to_string()));
+    /// assert_eq!(path.join("foo/bar")?.extension(), None);
+    ///
+    /// # Ok::<(), VfsError>(())
     pub fn extension(&self) -> Option<String> {
         let filename = self.filename();
         let mut parts = filename.rsplitn(2, '.');
@@ -291,6 +561,17 @@ impl VfsPath {
     /// Returns the parent path of this portion of this path
     ///
     /// Returns `None` if this is a root path
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsFileType, VfsMetadata, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    ///
+    /// assert_eq!(path.parent(), None);
+    /// assert_eq!(path.join("foo/bar")?.parent(), Some(path.join("foo")?));
+    /// assert_eq!(path.join("foo")?.parent(), Some(path));
+    ///
+    /// # Ok::<(), VfsError>(())
     pub fn parent(&self) -> Option<Self> {
         let index = self.path.rfind('/');
         index.map(|idx| VfsPath {
@@ -307,6 +588,21 @@ impl VfsPath {
     /// The returned paths may also point to non-existant files if there is concurrent removal.
     ///
     /// Also note that loops in the file system hierarchy may cause this iterator to never terminate.
+    ///
+    /// ```
+    /// # use vfs::{MemoryFS, VfsError, VfsPath, VfsResult};
+    /// let root = VfsPath::new(MemoryFS::new());
+    /// root.join("foo/bar")?.create_dir_all()?;
+    /// root.join("fizz/buzz")?.create_dir_all()?;
+    /// root.join("foo/bar/baz")?.create_file()?;
+    ///
+    /// let mut directories = root.walk_dir()?.collect::<VfsResult<Vec<_>>>()?;
+    ///
+    /// directories.sort_by_key(|path| path.as_str().to_string());
+    /// let expected = vec!["fizz", "fizz/buzz", "foo", "foo/bar", "foo/bar/baz"].iter().map(|path| root.join(path)).collect::<VfsResult<Vec<_>>>()?;
+    /// assert_eq!(directories, expected);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn walk_dir(&self) -> VfsResult<WalkDirIterator> {
         Ok(WalkDirIterator {
             inner: Box::new(self.read_dir()?),
@@ -317,6 +613,19 @@ impl VfsPath {
     /// Reads a complete file to a string
     ///
     /// Returns an error if the file does not exist or is not valid UTF-8
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let file = path.join("foo.txt")?;
+    /// write!(file.create_file()?, "Hello, world!")?;
+    ///
+    /// let result = file.read_to_string()?;
+    ///
+    /// assert_eq!(&result, "Hello, world!");
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn read_to_string(&self) -> VfsResult<String> {
         let metadata = self.metadata()?;
         if metadata.file_type != VfsFileType::File {
@@ -335,6 +644,20 @@ impl VfsPath {
     /// Copies a file to a new destination
     ///
     /// The destination must not exist, but its parent directory must
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let src = path.join("foo.txt")?;
+    /// write!(src.create_file()?, "Hello, world!")?;
+    /// let dest = path.join("bar.txt")?;
+    ///
+    /// src.copy_file(&dest)?;
+    ///
+    /// assert_eq!(dest.read_to_string()?, "Hello, world!");
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn copy_file(&self, destination: &VfsPath) -> VfsResult<()> {
         || -> VfsResult<()> {
             if destination.exists()? {
@@ -366,6 +689,21 @@ impl VfsPath {
     /// Moves or renames a file to a new destination
     ///
     /// The destination must not exist, but its parent directory must
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let src = path.join("foo.txt")?;
+    /// write!(src.create_file()?, "Hello, world!")?;
+    /// let dest = path.join("bar.txt")?;
+    ///
+    /// src.move_file(&dest)?;
+    ///
+    /// assert_eq!(dest.read_to_string()?, "Hello, world!");
+    /// assert!(!src.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn move_file(&self, destination: &VfsPath) -> VfsResult<()> {
         || -> VfsResult<()> {
             if destination.exists()? {
@@ -400,6 +738,20 @@ impl VfsPath {
     /// The destination must not exist, but the parent directory must
     ///
     /// Returns the number of files copied
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let src = path.join("foo")?;
+    /// src.join("dir")?.create_dir_all()?;
+    /// let dest = path.join("bar.txt")?;
+    ///
+    /// src.copy_dir(&dest)?;
+    ///
+    /// assert!(dest.join("dir")?.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn copy_dir(&self, destination: &VfsPath) -> VfsResult<u64> {
         let mut files_copied = 0u64;
         || -> VfsResult<()> {
@@ -430,31 +782,24 @@ impl VfsPath {
         Ok(files_copied)
     }
 
-    /// Returns `true` if the path exists and is pointing at a regular file, otherwise returns `false`.
-    ///
-    /// Note that this call may fail if the file's existence cannot be determined or the metadata can not be retrieved
-    pub fn is_file(&self) -> VfsResult<bool> {
-        if !self.exists()? {
-            return Ok(false);
-        }
-        let metadata = self.metadata()?;
-        Ok(metadata.file_type == VfsFileType::File)
-    }
-
-    /// Returns `true` if the path exists and is pointing at a directory, otherwise returns `false`.
-    ///
-    /// Note that this call may fail if the directory's existence cannot be determined or the metadata can not be retrieved
-    pub fn is_dir(&self) -> VfsResult<bool> {
-        if !self.exists()? {
-            return Ok(false);
-        }
-        let metadata = self.metadata()?;
-        Ok(metadata.file_type == VfsFileType::Directory)
-    }
-
     /// Moves a directory to a new destination, including subdirectories and files
     ///
     /// The destination must not exist, but its parent directory must
+    ///
+    /// ```
+    /// # use std::io::Read;
+    /// use vfs::{MemoryFS, VfsError, VfsPath};
+    /// let path = VfsPath::new(MemoryFS::new());
+    /// let src = path.join("foo")?;
+    /// src.join("dir")?.create_dir_all()?;
+    /// let dest = path.join("bar.txt")?;
+    ///
+    /// src.move_dir(&dest)?;
+    ///
+    /// assert!(dest.join("dir")?.exists()?);
+    /// assert!(!src.join("dir")?.exists()?);
+    /// # Ok::<(), VfsError>(())
+    /// ```
     pub fn move_dir(&self, destination: &VfsPath) -> VfsResult<()> {
         || -> VfsResult<()> {
             if destination.exists()? {
